# -*- coding: utf-8 -*-
# Copyright (C) 2016-2018 Cordell Bloor
# Published under the MIT License

"""Shared utilities between checkguard, guard2once and once2guard."""

from __future__ import print_function
import sys
import os
import re
from fnmatch import fnmatch
from functools import partial
from itertools import islice

py2 = sys.version_info < (3,)

def ends_with_blank_line(contents):
    """
    Returns true if the given string ends with a line that is either empty or
    only composed of whitespace.
    """
    return re.search('\n\s*\n\Z', contents) is not None

def next_nonempty_line(contents, start):
    """
    Returns the index of the first character of the next line containing
    non-whitespace characters, or returns len(contents) if there is no such
    line.
    """
    line_end = len(contents) - 1
    for i, c in enumerate(islice(contents, start, None), start):
        if c == '\n':
           line_end = i
        elif c not in ' \t':
           break
    return line_end + 1

def next_line(contents, start):
    """
    Returns the index of the first character of the next line, or returns
    len(contents) if there is no next line.
    """
    for i, c in enumerate(islice(contents, start, None), start):
        if c == '\n':
           return i + 1
    return len(contents)

def make_regex(pattern, guard_symbol=None):
    """
    Performs a few substitutions so the pattern can use format-style
    shorthands, then compiles and returns the regex.
    """
    replacements = {
        's' : r"[ \t]", # space characters
        'guard' : guard_symbol
    }
    return re.compile(pattern.format(**replacements), re.MULTILINE)

def make_regexes(patterns, guard_symbol=None):
    """
    Yields a generator that lazily compiles the given regex patterns.
    """
    for pattern in patterns:
        yield make_regex(pattern, guard_symbol)

def guess_guard(contents):
    """
    Returns the guard, as well as the start and end indexes of the include
    guard from the start of the file, or throws ValueError if not found.
    Comments are not supported.
    """
    patterns = [
        r"^{s}*[#]{s}*ifndef{s}+(?P<guard>[\w]+){s}*\n"
            + r"{s}*[#]{s}*define{s}+(?P=guard)({s}+1)?{s}*$",
        r"^{s}*[#]{s}*if{s}*[!]{s}*defined(?:{s}+|{s}*(?P<paren>[(]){s}*)(?P<guard>[\w]+){s}*(?(paren)[)]){s}*\n"
            + r"{s}*[#]{s}*define{s}+(?P=guard)({s}+1)?{s}*$",
    ]
    for regex in make_regexes(patterns):
        for match in regex.finditer(contents):
            # if it matches our regex, it's probably a guard, but check
            # that there's some content within to be a little more sure
            nextline_start = next_nonempty_line(contents, match.end())
            nextline_end = next_line(contents, nextline_start)
            try:
                index_guard_end(contents, nextline_start, nextline_end)
            except ValueError as e:
                return (match.group('guard'),) + match.span()
    raise ValueError('guard start not found')

def index_pragma_once(contents):
    """
    Returns the start and end indexes of the pragma once directive from
    the start of the file, or throws ValueError if not found. Comments
    are not supported.
    """
    regex = make_regex(r"^{s}*\#{s}*pragma{s}+once{s}*$")
    match = regex.search(contents)
    if match:
        return match.span()
    raise ValueError('pragma once not found')

def index_guard_start(contents, guard_symbol):
    """
    Returns the start and end indexes of the include guard from the start of the
    file, or throws ValueError if not found. Comments are not supported.
    """
    patterns = [
        r"^{s}*[#]{s}*ifndef{s}+{guard}{s}*\n"
            + r"{s}*[#]{s}*define{s}+{guard}(?:{s}+1)?{s}*$",
        r"^{s}*[#]{s}*if{s}*[!]{s}*defined(?:{s}+|{s}*(?P<paren>[(]){s}*){guard}{s}*(?(paren)[)]){s}*\n"
            + r"{s}*[#]{s}*define{s}+{guard}({s}+1)?{s}*$",
    ]
    for regex in make_regexes(patterns, guard_symbol):
        match = regex.search(contents)
        if match:
            return match.span()
    raise ValueError('guard start not found')

def index_guard_end(contents, pos=0, endpos=None):
    """
    Returns the start and end indexes of the last endif line from the
    file, or throws ValueError if not found. Comments are not supported.
    """
    if endpos is None:
        endpos = len(contents)
    regex = make_regex(r"^{s}*[#]{s}*endif{s}*(?:[/][/*].*)?$")
    match = None
    for match in regex.finditer(contents, pos, endpos):
        pass
    if match:
        return match.span()
    raise ValueError('guard end not found')

class FileMetadata:
    """
    Information about a file's encoding and line endings.
    """
    def __init__(self):
        self.bom = False
        self.crlf = False

bom = b'\xef\xbb\xbf'

def ds(s):
    """
    Decodes the given byte string in Python 3.
    """
    if py2:
        return s
    return s.decode(encoding='utf-8', errors='surrogateescape')

def es(s):
    """
    Encodes the given string in Python 3.
    """
    if py2:
        return s
    return s.encode(encoding='utf-8', errors='surrogateescape')

def process_input(contents):
    """
    Analyse file contents and simplify. Strip the BOM if necessary,
    normalize line endings, and decode.
    """
    metadata = FileMetadata()
    if contents.startswith(bom):
        contents = contents[len(bom):]
        metadata.bom = True
    first_newline_idx = contents.find(b'\n')
    if first_newline_idx > 0:
        metadata.crlf = contents.startswith(b'\r', first_newline_idx-1)
        contents = contents.replace(b'\r\n', b'\n')
    contents = ds(contents)
    return (contents, metadata);

def get_file_contents(filename):
    if filename is None:
        return process_input(sys.stdin.read())
    with open(filename, 'rb') as f:
        return process_input(f.read())

def process_output(contents, metadata):
    """
    Undo the operations done in process_input to prepare contents for writing.
    """
    contents = es(contents)
    if metadata.crlf:
        contents = contents.replace(b'\n', b'\r\n')
    if metadata.bom:
        contents = bom + contents
    return contents

def write_file_contents(filename, contents, metadata):
    if filename is None:
        return sys.stdout.write(ds(process_output(contents, metadata)))
    with open(filename, 'wb') as f:
        return f.write(process_output(contents, metadata))

def is_header_file(filename):
    '''Returns true if the given file is identified as a C/C++ header file.'''
    return filename.endswith(('.h', '.hpp', '.H', '.hh', '.hxx'))

def is_excluded(filepath, exclusions):
    '''Returns true if the given file matches any of the exclusion patterns.'''
    return any(map(partial(fnmatch, filepath), exclusions))

def apply_to_headers(func, directory, exclusions):
    '''
    Apply the given function to all the header files in the given directory and
    all its subdirectories, excluding files and subdirectories matching any
    pattern in the list of exclusions.
    '''
    # Workaround for lack of nonlocal keyword in Python 2.7
    class Status:
        pass
    status = Status()
    status.ok = True

    def report_error(error):
        '''Report an error to the user, but continue processing.'''
        status.ok = False
        print(error, file=sys.stderr)

    for root, dirs, files in os.walk(directory, onerror=report_error):
        for filename in files:
            filepath = os.path.join(root, filename)
            if is_header_file(filename) and not is_excluded(filepath, exclusions):
                status.ok &= func(filepath, filename)
    return status.ok



# -*- coding: utf-8 -*-
# Copyright (C) 2017 Cordell Bloor
# Published under the MIT License

class Template:
    """
    A purpose-built string formatter. It replaces % with the value passed to sub.
    % can be escaped as %%. It does nothing else, so there are no surprises.
    string.Template is a decent alternative, but its syntax is more complex
    than really necessary for these purposes.
    """
    def __init__(self, fmtstr):
        self.placeholder = object()
        self.pieces = gather_pieces(fmtstr, self.placeholder)

    def sub(self, guard):
        return ''.join([guard if x is self.placeholder else x for x in self.pieces])

def gather_pieces(fmtstr, placeholder):
    """
    Takes a format string where % marks replacements by placeholder, and %%
    marks replacements by %. The returned object is a list of substrings and
    placeholders.
    """
    pieces = []
    substr = []
    escape = False
    for c in fmtstr:
        if escape:
            if c == '%':
                substr.append(c)
            else:
                pieces.append(''.join(substr))
                substr = []
                pieces.append(placeholder)
                substr.append(c)
            escape = False
        else:
            if c == '%':
                escape = True
            else:
                substr.append(c)
    pieces.append(''.join(substr))
    if escape:
        pieces.append(placeholder)
    return pieces



# -*- coding: utf-8 -*-
# Copyright (C) 2016-2018 Cordell Bloor
# Published under the MIT License

"""Replace #pragma once with C and C++ include guards."""

from __future__ import print_function
import argparse
import sys
import os
from functools import partial
from .pattern_compiler import compile_pattern, ParserError
from .util import (index_pragma_once, get_file_contents, write_file_contents,
    apply_to_headers, ends_with_blank_line, py2)
from .template import Template

__version__ = "2.4.0"

def replace_pragma_once(contents, guard,
        endif_template=Template('#endif\n'), endif_newline=False):
    guard_open = '#ifndef {0}\n#define {0}'.format(guard)
    guard_close = endif_template.sub(guard)
    try:
        once_start, once_end = index_pragma_once(contents)
    except ValueError:
        return None

    # figure out how many newlines to put before the guard
    if ends_with_blank_line(contents):
        nl = ''
    elif contents.endswith('\n'):
        nl = '\n' if endif_newline else ''
    else:
        nl = '\n\n' if endif_newline else '\n'

    return (contents[:once_start]
        + guard_open
        + contents[once_end:]
        + nl + guard_close)

def process_file(filepath, filename, options):
    class Context:
        pass
    ctx = Context()
    ctx.filepath = filepath
    ctx.filename = filename

    options.guard = options.create_guard(ctx)

    success = True
    try:
        contents, metadata = get_file_contents(filepath)
        new_contents = replace_pragma_once(contents, options.guard,
            options.endif_template, options.endif_newline)
        if options.stdout and new_contents is None:
            new_contents = contents
        if new_contents:
            output_path = None if options.stdout else filepath
            write_file_contents(output_path, new_contents, metadata)
    except Exception as e:
        success = False
        print('Error processing {0}:\n\t({1}) {2}'.format(filepath,
            e.__class__.__name__, str(e)), file=sys.stderr)
    return success

def process_guard_pattern(pattern):
    create_guard = lambda ctx: None
    if pattern is not None:
        try:
            create_guard = compile_pattern(pattern)
        except ParserError as e:
            print(e, file=sys.stderr)
            sys.exit(1)
    return create_guard

def eus(ustr):
    '''
    Encodes the given unicode string to str
    '''
    if py2:
        return ustr.encode('utf8')
    else:
        return ustr

def decode_escapes(s):
    import re
    import codecs
    def decode_match(match):
        return eus(codecs.decode(match.group(0), 'unicode-escape'))
    # https://stackoverflow.com/a/24519338
    escapes = re.compile(r'''
        ( \\U........      # 8-digit hex escapes
        | \\u....          # 4-digit hex escapes
        | \\x..            # 2-digit hex escapes
        | \\[0-7]{1,3}     # Octal escapes
        | \\N\{[^}]+\}     # Unicode characters by name
        | \\[\\'"abfnrtv]  # Single-character escapes
        )''', re.UNICODE | re.VERBOSE)
    return escapes.sub(decode_match, s)

def main():
    parser = argparse.ArgumentParser(
            description='Replace #pragma once with C and C++ include guards.')
    parser.add_argument('files',
            metavar='file',
            nargs='+',
            help='the file(s) to check; directories require the recursive '
            'option')
    parser.add_argument('-V','--version', action='version',
            version='%(prog)s ' + __version__)
    parser.add_argument('-r','--recursive',
            action='store_true',
            dest='recursive',
            help='recursively search directories for headers')
    parser.add_argument('-s','--endif-style',
            dest='endif_template',
            metavar='template',
            default='#endif\n',
            help='use the given template for the inserted #endif. '
            'The include guard can be referenced in the template with %%. '
            'Any other uses of %% must be escaped as %%%%.')
    parser.add_argument('-l','--endif-newline',
            action='store_true',
            dest='endif_newline',
            help='insert a blank line before #endif if none would exist '
            'otherwise')
    parser.add_argument('-p','--pattern',
            default='name|upper',
            metavar='pattern',
            help='generate include guards that match the specified pattern. '
            "For example, -p 'name | upper' would create the guard MATCH_H "
            'for the file Match.h. See the docs on GitHub for a full '
            'description of the guard pattern language.')
    parser.add_argument('-e','--exclude',
            action='append',
            dest='exclusions',
            metavar='pattern',
            default=[],
            help='exclude files that match the given fnmatch pattern. '
            'Any * is a wildcard matching everything; '
            'a ? matches any single character; '
            '[_] matches any characters within the brackets; '
            'and [!_] matches any characters not within the brackets.')
    parser.add_argument('--stdout',
            action='store_true',
            dest='stdout',
            help='write output to stdout')
    args = parser.parse_args()

    class Options:
        pass
    options = Options()
    options.create_guard = process_guard_pattern(args.pattern)
    options.endif_template = Template(decode_escapes(args.endif_template))
    options.endif_newline = args.endif_newline
    options.stdout = args.stdout

    if options.stdout and args.recursive:
        print('The recursive option cannot be used when printing to stdout',
            file=sys.stderr)
        sys.exit(1)
    elif options.stdout and len(args.files) > 1:
        print('Only one file can be specified at a time when printing to stdout',
            file=sys.stderr)
        sys.exit(1)

    ok = True
    for f in args.files:
        if os.path.isdir(f):
            if args.recursive:
                process = partial(process_file, options=options)
                ok &= apply_to_headers(process, f, args.exclusions)
            else:
                print('"%s" is a directory' % f, file=sys.stderr)
                sys.exit(1)
        else:
            ok &= process_file(f, os.path.basename(f), options)
    if not ok:
        sys.exit(1)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright (C) 2016-2017 Cordell Bloor
# Published under the MIT License

"""Create functions that generate include guard tokens from patterns."""

import os
import re
from string import capwords

class ParseState:
    Normal, Token, SingleQuote, DoubleQuote, Complete = range(5)

class ParserError(Exception):
    pass

def next_token(pattern, start):
    """
    Given a pattern string and a start index, returns the end index of the next
    token, including whitespace on either side.
    """
    state = ParseState.Normal
    index = start
    for i in range(start, len(pattern)):
        if state == ParseState.SingleQuote:
            if pattern[i] == "'":
                state = ParseState.Complete
        elif state == ParseState.DoubleQuote:
            if pattern[i] == '"':
                state = ParseState.Complete
        else:
            if pattern[i].isspace():
                if state == ParseState.Token:
                    state = ParseState.Complete
            elif pattern[i] == '|':
                break # delimits tokens
            elif pattern[i] == '"':
                if state == ParseState.Complete:
                    break # start of new token
                state = ParseState.DoubleQuote
            elif pattern[i] == "'":
                if state == ParseState.Complete:
                    break # start of new token
                state = ParseState.SingleQuote
            else:
                if state == ParseState.Complete:
                    break # start of new token
                state = ParseState.Token
        index = i

    if state in [ParseState.SingleQuote, ParseState.DoubleQuote]:
        raise ParserError('Pattern contains unclosed quote')

    return index + 1

def unquote(s):
    """
    Strips matching single and double quotes from the start and end of the
    given string.
    """
    if len(s) > 1 and ((s[0] == '"' and s[-1] == '"') or
                       (s[0] == "'" and s[-1] == "'")):
        return s[1:-1]
    else:
        return s

def tokenize(pattern):
    tokens = []
    start = 0
    end = 0
    while end != len(pattern):
        end = next_token(pattern, start)
        tokens.append(pattern[start:end].strip())
        start = end
    return tokens

def trim_parents(filepath, crumbs):
    """
    Discards part of the given filepath. The base file is always kept. The
    crumbs argument specifies how many parent directories to keep. Negative
    crumb values are interpreted as relative to the total number of parents.
    """
    if crumbs < 0:
        if os.path.isabs(filepath):
            crumbs -= 1
        crumbs = max(crumbs + filepath.count(os.sep), 0)
    idx = len(filepath)
    for i in range(crumbs + 1):
        if idx >= 0:
            idx = filepath.rfind(os.sep, 0, idx)
    if idx < 0:
        return filepath
    else:
        return filepath[idx+1:]

def snake(s):
    """
    Converts an input string in PascalCase to snake_case.
    """
    snek = []
    prev_up = False
    prev_alnum = False
    for idx, c in enumerate(s):
        alnum = c.isalnum()
        up = c.isupper()
        next_up = s[idx+1].isupper() if idx+1 < len(s) else False
        if (up and not prev_up and prev_alnum and idx != 0 or
            up and prev_up and not next_up and idx != len(s)):
            snek.append('_')
        snek.append(c.lower())
        prev_alnum = alnum
        prev_up = up
    return ''.join(snek)

def pascal(s):
    """
    Converts an input string in snake_case to PascalCase. Also handles
    names containing spaces... god help you if you actually need that.
    """
    return capwords(s.replace('_',' ')).replace(' ','')

def sanitize(s):
    """
    Removes characters that are not allowed in macro names. Anything
    that's not alphanumeric is replaced with underscore.
    """
    return re.sub(r"\W", '_', s)

class Args:
    """
    An enum representing all arguments that can be passed to functions.
    """
    (Replace, ReplaceWith, AppendWith, PrependWith, SurroundWith,
        PathCrumbs, Remove, TrimParentsBy) = range(1,9)

def compile_pattern(pattern):
    """
    Takes a pattern specification as a string, and returns a function that
    turns a file context into an include guard.
    """
    sources = ['name','path']

    chain = []
    function = None
    expected_arg = None
    optional_arg = False
    args = []
    raw = False

    tokens = tokenize(pattern)
    if not tokens:
        raise ParserError('Pattern is empty')
    elif tokens[0] not in sources:
        raise ParserError('First function in pattern must be a source')
    elif tokens[-1] == '|':
        raise ParserError('Pattern ends with pipe to nowhere')

    for token in tokens:
        if raw:
            raise ParserError('Raw function is a sink and must be last')

        if not expected_arg:
            if not function:
                function = token
            elif token == '|':
                function = None
            else:
                raise ParserError('Unexpected argument "%s" in pattern' % token)

            if token == 'name':
                chain.append(lambda ctx, s: ctx.filename)
            elif token == 'path':
                expected_arg = Args.PathCrumbs
                optional_arg = True
            elif token == 'upper':
                chain.append(lambda ctx, s: s.upper())
            elif token == 'lower':
                chain.append(lambda ctx, s: s.lower())
            elif token == 'snake':
                chain.append(lambda ctx, s: snake(s))
            elif token == 'pascal':
                chain.append(lambda ctx, s: pascal(s))
            elif token == 'remove':
                expected_arg = Args.Remove
            elif token == 'replace':
                expected_arg = Args.Replace
            elif token == 'append':
                expected_arg = Args.AppendWith
            elif token == 'prepend':
                expected_arg = Args.PrependWith
            elif token == 'surround':
                expected_arg = Args.SurroundWith
            elif token == 'parents':
                expected_arg = Args.TrimParentsBy
            elif token == 'raw':
                raw = True
            elif token != '|':
                raise ParserError('Unknown function "%s" in pattern' % token)
        elif token == '|':
            if optional_arg:
                function = None
                if expected_arg == Args.PathCrumbs:
                    chain.append(lambda ctx, s: ctx.filepath)
                    expected_arg = None
                    optional_arg = False
            else:
                raise ParserError('Missing argument from "%s" in pattern' % function)
        elif expected_arg == Args.Remove:
            chain.append(lambda ctx, s, t=unquote(token): s.replace(t, ''))
            expected_arg = None
        elif expected_arg == Args.Replace:
            expected_arg = Args.ReplaceWith
            args.append(unquote(token))
        elif expected_arg == Args.ReplaceWith:
            chain.append(lambda ctx, s, f=args[0], t=unquote(token): s.replace(f, t))
            expected_arg = None
            args = []
        elif expected_arg == Args.AppendWith:
            chain.append(lambda ctx, s, suffix=unquote(token): s + suffix)
            expected_arg = None
        elif expected_arg == Args.PrependWith:
            chain.append(lambda ctx, s, prefix=unquote(token): prefix + s)
            expected_arg = None
        elif expected_arg == Args.SurroundWith:
            chain.append(lambda ctx, s, arg=unquote(token): arg + s + arg)
            expected_arg = None
        elif expected_arg == Args.TrimParentsBy:
            try:
                crumbs = int(token)
            except ValueError:
                raise ParserError('Invalid argument "%s" given to parents must be an integer' % token)
            chain.append(lambda ctx, s, crumbs=crumbs: trim_parents(s, crumbs))
            expected_arg = None
        elif expected_arg == Args.PathCrumbs:
            try:
                crumbs = int(token)
            except ValueError:
                raise ParserError('Invalid argument "%s" given to path; must be an integer' % token)
            chain.append(lambda ctx, s, crumbs=crumbs: trim_parents(ctx.filepath, crumbs))
            expected_arg = None
            optional_arg = False

    if optional_arg:
        if expected_arg == Args.PathCrumbs:
            chain.append(lambda ctx, s: ctx.filepath)
    elif expected_arg:
        raise ParserError('Missing argument from "%s" in pattern' % function)

    def process(ctx):
        s = ''
        for fn in chain:
            s = fn(ctx, s)
        return s if raw else sanitize(s)
    return process
